@@ -1,24 +1,28 @@
 """
 Ludolph: Monitoring Jabber Bot
-Copyright (C) 2012-2013 Erigones s. r. o.
+Copyright (C) 2012-2014 Erigones s. r. o.
 This file is part of Ludolph.
 
 See the LICENSE file for copying permission.
 """
 
-import os
 import ssl
 import time
 import logging
 from sleekxmpp import ClientXMPP
 from sleekxmpp.xmlstream import ET
+from sleekxmpp.exceptions import IqError
 
 from ludolph.message import LudolphMessage
 from ludolph.command import COMMANDS, USERS, ADMINS
+from ludolph.web import WebServer
+from ludolph.cron import Cron
 
 logger = logging.getLogger(__name__)
 
-__all__ = ['LudolphBot']
+__all__ = ('LudolphBot',)
+
+PLUGINS = {}  # {modname : instance}
 
 
 class LudolphBot(ClientXMPP):
@@ -26,10 +30,9 @@ class LudolphBot(ClientXMPP):
     Ludolph bot.
     """
     _start_time = None
-    _commands = None  # Cached sorted list of commands
     _muc_ready = False
     commands = COMMANDS
-    plugins = None
+    plugins = PLUGINS
     users = USERS
     admins = ADMINS
     room = None
@@ -38,14 +41,16 @@ class LudolphBot(ClientXMPP):
     room_admins = set()
     muc = None
     nick = 'Ludolph'
-    pipe_file = None
     xmpp = None
-    maxhistory = '1024'
+    maxhistory = '4096'
+    webserver = None
+    cron = None
 
+    # noinspection PyUnusedLocal
     def __init__(self, config, plugins=None, *args, **kwargs):
-        self._load_config(config)
+        self._load_config(config, init=True)
         logger.info('Initializing *%s* jabber bot', self.nick)
-        self._load_plugins(config, plugins)
+        self._load_plugins(config, plugins, init=True)
 
         # Initialize the SleekXMPP client
         ClientXMPP.__init__(self, config.get('xmpp', 'username'), config.get('xmpp', 'password'))
@@ -68,21 +73,24 @@ class LudolphBot(ClientXMPP):
             self.add_event_handler('groupchat_message', self.muc_message, threaded=True)
             self.add_event_handler('muc::%s::got_online' % self.room, self.muc_online, threaded=True)
 
-        # Start the monitoring thread for reading the pipe file
-        self._start_thread('mon_thread', self.mon_thread)
+        # Start the web server thread for processing HTTP requests
+        if self.webserver:
+            self._start_thread('webserver', self.webserver.start, track=False)
+
+        # Start the scheduler thread for running periodic cron jobs
+        if self.cron:
+            self._start_thread('cron', self.cron.run, track=False)
 
         # Save start time
         self._start_time = time.time()
 
-    def _load_config(self, config):
+    def _load_config(self, config, init=False):
         """
         Load bot settings from config object.
+        The init parameter indicates whether this is a first-time initialization or a reload.
         """
         logger.info('Configuring jabber bot')
 
-        # Global stuff
-        self.pipe_file = config.get('global', 'pipe_file')
-
         # Get nick name
         if config.has_option('xmpp', 'nick'):
             nick = config.get('xmpp', 'nick').strip()
@@ -157,7 +165,21 @@ class LudolphBot(ClientXMPP):
                 logger.error('Room admin "%s" is not specified in room_users. '
                              'This may lead to unexpected behaviour. ', i)
 
-    def _load_plugins(self, config, plugins, init=True):
+        # Web server (any change in configuration requires restart)
+        if init and not self.webserver:
+            if config.has_option('webserver', 'host') and config.has_option('webserver', 'port'):
+                host = config.get('webserver', 'host').strip()
+                port = config.getint('webserver', 'port')
+
+                if host and port:  # Enable server (will be started in __init__)
+                    self.webserver = WebServer(host, port)
+
+        # Cron (any change in configuration requires restart)
+        if init and not self.cron:
+            if config.has_option('cron', 'enabled') and config.getboolean('cron', 'enabled'):
+                self.cron = Cron()
+
+    def _load_plugins(self, config, plugins, init=False):
         """
         Initialize plugins.
         The init parameter indicates whether this is a first-time initialization or a reload.
@@ -165,31 +187,58 @@ class LudolphBot(ClientXMPP):
         if init:
             # First-time plugin initialization -> include ourself to plugins dict
             self.xmpp = self
-            self.plugins = {__name__: self}
+            self.plugins.clear()
+            self.plugins[__name__] = self
         else:
             # Bot reload - remove disabled plugins
-            for enabled_plugin in self.plugins.keys():
+            for enabled_plugin in tuple(self.plugins.keys()):  # Copy for python 3
                 if enabled_plugin == __name__:
                     continue  # Skip ourself
 
-                if not plugins or enabled_plugin not in plugins:
-                    logger.info('Disabling plugin %s', enabled_plugin)
+                if enabled_plugin not in plugins:
+                    logger.info('Disabling plugin: %s', enabled_plugin)
                     del self.plugins[enabled_plugin]
 
         if plugins:
-            for plugin, cls in plugins.items():
-                if init or plugin not in self.plugins:
-                    logger.info('Initializing plugin %s', plugin)
-                    self.plugins[plugin] = cls(config, reinit=False)
+            for modname, plugin in plugins.items():
+                if init or modname not in self.plugins:
+                    logger.info('Initializing plugin: %s', modname)
+                    reinit = False
                 else:
-                    logger.info('Reloading plugin %s', plugin)
-                    del self.plugins[plugin]
-                    self.plugins[plugin] = cls(config, reinit=True)
-                # xmpp attribute pointing to this instance is available in plugin object
-                setattr(self.plugins[plugin], 'xmpp', self)
+                    logger.info('Reloading plugin: %s', modname)
+                    del self.plugins[modname]
+                    reinit = True
 
-        cmds = self.available_commands(reset=True)
-        logger.info('Registered commands: %s', ', '.join(cmds))
+                try:
+                    cfg = config.items(plugin[0])  # Get only plugin config section as list of (name, value) tuples
+                    obj = plugin[1](self, cfg, reinit=reinit)
+                except Exception as ex:
+                    logger.critical('Could not load plugin: %s', modname)
+                    logger.exception(ex)
+                else:
+                    self.plugins[modname] = obj
+
+        # Update commands cache
+        if self.commands.all(reset=True):
+            logger.info('Registered commands:\n%s\n', '\n'.join(self.commands.display()))
+        else:
+            logger.warning('NO commands registered')
+
+        if self.webserver:
+            if self.webserver.webhooks:
+                logger.info('Registered webhooks:\n%s\n', '\n'.join(self.webserver.display_webhooks()))
+            else:
+                logger.warning('NO webhooks registered')
+        else:
+            logger.warning('Web server support disabled - webhooks will not work')
+
+        if self.cron:
+            if self.cron.crontab:
+                logger.info('Registered cron jobs:\n%s\n', '\n'.join(self.cron.display_cronjobs()))
+            else:
+                logger.warning('NO cron jobs registered')
+        else:
+            logger.warning('Cron support disabled - cron jobs will not work')
 
     def _room_members(self):
         """
@@ -233,9 +282,18 @@ class LudolphBot(ClientXMPP):
             self.room_config['fields']['members_by_default']['value'] = True
 
         logger.info('Setting new configuration for MUC room %s', self.room)
-        self.muc.setRoomConfig(self.room, self.room_config)
+        try:
+            self.muc.setRoomConfig(self.room, self.room_config)
+        except IqError as e:
+            logger.error('Could not configure MUC room. Error was: %s (condition=%s, etype=%s)',
+                         e.text, e.condition, e.etype)
+
         logger.info('Setting member list for MUC room %s', self.room)
-        self._room_members()
+        try:
+            self._room_members()
+        except IqError as e:
+            logger.error('Could not configure MUC room member list. Error was: %s (condition=%s, etype=%s)',
+                         e.text, e.condition, e.etype)
 
     def _jid_in_room(self, jid):
         """
@@ -277,6 +335,7 @@ class LudolphBot(ClientXMPP):
 
         return jid
 
+    # noinspection PyUnusedLocal
     def session_start(self, event):
         """
         Process the session_start event.
@@ -298,7 +357,7 @@ class LudolphBot(ClientXMPP):
 
         # Remove users with none subscription from roster
         # Also remove users that are not in users setting (if set)
-        for i in roster.keys():
+        for i in roster:
             if roster[i]['subscription'] == 'none' or (self.users and i not in self.users):
                 logger.warning('Roster item: %s (%s) - removing!', i, roster[i]['subscription'])
                 self.send_presence(pto=i, ptype='unsubscribe')
@@ -333,45 +392,6 @@ class LudolphBot(ClientXMPP):
                         muc['jid'], muc['nick'], muc['role'], muc['affiliation'])
             self.msg_send(presence['from'].bare, 'Hello %s!' % muc['nick'], mtype='groupchat')
 
-    def available_commands(self, reset=False):
-        """
-        List of all available bot commands.
-        """
-        # Remove commands from disabled plugins
-        if reset:
-            for cmd_name, cmd in self.commands.items():
-                if cmd['module'] not in self.plugins:
-                    del self.commands[cmd_name]
-            self._commands = None
-
-        # Sort and cache
-        if self._commands is None:
-            self._commands = sorted(self.commands.keys())
-
-        return self._commands
-
-    def get_command(self, cmdstr):
-        """
-        Find text in available commands and return command dict.
-        """
-        if not cmdstr:
-            return None
-
-        if cmdstr in self.available_commands():
-            cmd = self.commands[cmdstr]
-        else:
-            for key in self.available_commands():
-                if key.startswith(cmdstr):
-                    cmd = self.commands[key]
-                    cmdstr = key
-                    break
-            else:
-                return None
-
-        cmd['str'] = cmdstr
-
-        return cmd
-
     def message(self, msg, types=('chat', 'normal')):
         """
         Incoming message handler.
@@ -380,17 +400,17 @@ class LudolphBot(ClientXMPP):
             return
 
         # Seek received text in available commands and get command
-        cmd = self.get_command(msg['body'].split()[0].strip())
+        cmd = self.commands.get_command(msg['body'].split()[0].strip())
 
         if cmd:
             start_time = time.time()
-            f_cmd = getattr(self.plugins[cmd['module']], cmd['name'])
+            f_cmd = getattr(self.plugins[cmd[1]], cmd[0])  # Get command bound method from plugin
             # Run command
             out = f_cmd(msg)
 
             if out:
                 cmd_time = time.time() - start_time
-                logger.info('Command %s.%s finished in %g seconds', cmd['module'], cmd['name'], cmd_time)
+                logger.info('Command %s.%s finished in %g seconds', cmd[1], cmd[0], cmd_time)
 
             return out
         else:
@@ -415,20 +435,40 @@ class LudolphBot(ClientXMPP):
             msg['body'] = msg['body'].lstrip(nick).lstrip()
             return self.message(msg, types=('groupchat',))
 
+    # noinspection PyUnusedLocal
     def shutdown(self, signalnum, handler):
         """
         Shutdown signal handler.
         """
         logger.info('Requested shutdown (%s)', signalnum)
 
-        return self.abort()
+        if self.webserver:
+            self.webserver.stop()
+
+        if self.cron:
+            self.cron.stop()
+
+        self.abort()
+
+    def prereload(self):
+        """
+        Cleanup during reload phase. Runs before plugin loading in main.
+        """
+        self.commands.reset()
+
+        if self.webserver:
+            self.webserver.reset_webhooks()
+            self.webserver.reset_webapp()
+
+        if self.cron:
+            self.cron.reset()
 
     def reload(self, config, plugins=None):
         """
         Reload bot configuration and plugins.
         """
         logger.info('Requested reload')
-        self._load_config(config)
+        self._load_config(config, init=False)
         self._load_plugins(config, plugins, init=False)
 
         if self.room and self.muc:
@@ -437,42 +477,13 @@ class LudolphBot(ClientXMPP):
             logger.info('Reinitializing multi-user chat room %s', self.room)
             self.muc.joinMUC(self.room, self.nick, maxhistory=self.maxhistory)
 
-    def mon_thread(self):
-        """
-        Processing input from the monitoring pipe file.
-        """
-        with os.fdopen(os.open(self.pipe_file, os.O_RDONLY | os.O_NONBLOCK)) as fifo:
-            logger.info('Processing input from monitoring pipe file')
-            while not self.stop.is_set():
-                line = fifo.readline().strip()
-                if line:
-                    data = line.split(';', 1)
-                    if len(data) == 2:
-                        logger.info('Sending monitoring message to "%s"', data[0])
-                        logger.debug('\twith body: "%s"', data[1])
-
-                        if data[0] == self.room:
-                            mtype = 'groupchat'
-                        else:
-                            mtype = 'normal'
-
-                        self.msg_send(data[0], data[1], mtype=mtype)
-
-                    else:
-                        logger.warning('Bad message format ("%s")', line)
-
-                time.sleep(1)
-
-                if self.stop.is_set():
-                    self._end_thread('mon_thread', early=True)
-                    return
-
     def msg_send(self, mto, mbody, **kwargs):
         """
         Create message and send it.
         """
         return LudolphMessage.create(mbody, **kwargs).send(self, mto)
 
+    # noinspection PyMethodMayBeStatic
     def msg_reply(self, msg, mbody, **kwargs):
         """
         Set message reply text and html, and send it.
